Quantile_Mapping_fake <- function (exp, obs, exp_cor = NULL, sdate_dim = "sdate", memb_dim = "member", 
                                   window_dim = NULL, method = "QUANT", na.rm = FALSE, ncores = NULL, 
                                   ...) 
{
  obsdims <- names(dim(obs))
  expdims <- names(dim(exp))
  if (!is.array(exp) || !is.numeric(exp)) {
    stop("Parameter 'exp' must be a numeric array.")
  }
  if (!is.array(obs) || !is.numeric(obs)) {
    stop("Parameter 'obs' must be a numeric array.")
  }
  if (is.null(expdims)) {
    stop("Parameter 'exp' must have dimension names.")
  }
  if (is.null(obsdims)) {
    stop("Parameter 'obs' must have dimension names.")
  }
  if (!is.character(sdate_dim) | length(sdate_dim) != 1) {
    stop("Parameter 'sdate_dim' must be a character string.")
  }
  if (!sdate_dim %in% expdims | !sdate_dim %in% obsdims) {
    stop("Parameter 'sdate_dim' is not found in 'exp' or 'obs' dimension.")
  }
  if (dim(exp)[sdate_dim] == 1 || dim(obs)[sdate_dim] == 1) {
    stop("Parameter 'exp' and 'obs' must have dimension length of 'sdate_dim' bigger than 1.")
  }
  if (!is.null(exp_cor)) {
    if (is.null(names(dim(exp_cor)))) {
      stop("Parameter 'exp_cor' must have dimension names.")
    }
    if (!sdate_dim %in% names(dim(exp_cor))) {
      stop("Parameter 'sdate_dim' is not found in 'exp_cor' dimension.")
    }
  }
  if (!(method %in% c("PTF", "DIST", "RQUANT", "QUANT", "SSPLIN")) | 
      length(method) != 1) {
    stop("Parameter 'method' must be one of the following methods: ", 
         "'PTF', 'DIST', 'RQUANT', 'QUANT', 'SSPLIN'.")
  }


  if (is.null(memb_dim)) {
    remove_member <- TRUE
    memb_dim <- "temp_memb_dim"
    exp <- s2dv::InsertDim(exp, posdim = 1, lendim = 1, name = "temp_memb_dim")
    obs <- s2dv::InsertDim(obs, posdim = 1, lendim = 1, name = "temp_memb_dim")
    obsdims <- names(dim(obs))
    expdims <- names(dim(exp))
    if (!is.null(exp_cor)) {
      exp_cor <- s2dv::InsertDim(exp_cor, posdim = 1, lendim = 1, 
                           name = "temp_memb_dim")
    }
  }
  else {
    remove_member <- FALSE
    if (!all(memb_dim %in% obsdims)) {
      obs <- s2dv::InsertDim(obs, posdim = 1, lendim = 1, name = memb_dim[!(memb_dim %in% 
                                                                        obsdims)])
      obsdims <- names(dim(obs))
    }
    if (any(!memb_dim %in% expdims)) {
      stop(paste0("Parameter 'memb_dim' is not found in 'exp' dimensions. ", 
                  "Set it as NULL if there is no member dimension."))
    }
  }
  sample_dims <- c(memb_dim, sdate_dim)
  if (!is.null(window_dim)) {
    if (!(window_dim %in% obsdims)) {
      stop("Parameter 'window_dim' is not found in 'obs'.")
    }
    obs <- CSTools::MergeDims(obs, c(memb_dim, window_dim))
    if (window_dim %in% expdims) {
      exp <- CSTools::MergeDims(exp, c(memb_dim, window_dim))
      warning("Parameter 'window_dim' is found in exp and is merged to 'memb_dim'.")
    }
  }
  if (!is.logical(na.rm) | length(na.rm) > 1) {
    stop("Parameter 'na.rm' must be one logical value.")
  }
  if (!is.null(ncores)) {
    if (!is.numeric(ncores) | ncores%%1 != 0 | ncores <= 
        0 | length(ncores) > 1) {
      stop("Parameter 'ncores' must be either NULL or a positive integer.")
    }
  }
  if (!is.null(exp_cor)) {
    # browser()
    qmaped <- Apply(list(exp, obs, exp_cor), target_dims = sample_dims, 
                    fun = qmapcor, method = method, sdate_dim = sdate_dim, 
                    na.rm = na.rm, ..., ncores = 1)$output1
  }
  else {
    qmaped <- Apply(list(exp, obs), target_dims = sample_dims, 
                    fun = qmapcor, exp_cor = NULL, method = method, 
                    sdate_dim = sdate_dim, na.rm = na.rm, ..., ncores = ncores)$output1
  }
  if (remove_member) {
    dim(qmaped) <- dim(qmaped)[-which(names(dim(qmaped)) == 
                                        "temp_memb_dim")]
  }
  return(qmaped)
}

qmapcor <- function(exp, obs, exp_cor = NULL, sdate_dim = 'sdate', 
                     method = 'QUANT', na.rm = FALSE, ...) {
  
  # exp: [memb (+ window), sdate]
  # obs: [memb (+ window), sdate] 
  # exp_cor: NULL or [memb, sdate]
  if (is.null(exp_cor)) {
    applied <- exp * NA
    for (sd in 1:dim(exp)[sdate_dim]) {
      if (na.rm) {
        # select start date for cross-val
        nas_pos <- which(!is.na(exp[, sd]))
        obs2 <- as.vector(obs[, -sd])
        exp2 <- as.vector(exp[, -sd])
        exp_cor2 <- as.vector(exp[, sd])
        # remove NAs
        obs2 <- obs2[!is.na(obs2)]
        exp2 <- exp2[!is.na(exp2)]   
        exp_cor2 <- exp_cor2[!is.na(exp_cor2)]
        tryCatch({
          adjust <- fitQmap(obs2, exp2, method = method, ...)
          applied[nas_pos, sd] <- doQmap(exp_cor2, adjust, ...)
        },
        error = function(error_message) {
          return(applied[, sd])
        })
      } else {
        # na.rm = FALSE shouldn't fail, just return NA
        if (anyNA(obs[, -sd]) | anyNA(exp[, -sd])) {
          applied[, sd] <- NA
        } else {
          adjust <- fitQmap(as.vector(obs[, -sd]), as.vector(exp[, -sd]),
                            method = method, ...)
          exp2 <- exp[, sd]
          if (sum(is.na(exp2)) >= 1) {
            app <- rep(NA, length(exp2))
            nas_pos <- which(is.na(exp2))
            exp2 <- exp2[!is.na(exp2)]
            app[-nas_pos] <- doQmap(as.vector(exp2), adjust, ...)
          } else {
            app <- doQmap(as.vector(exp2), adjust, ...)
          }
          applied[, sd] <- app
        }
      }
    }
  } else {
    applied <- exp_cor * NA
    if (na.rm) {
      tryCatch({
        adjust <- fitQmap(obs[!is.na(obs)], exp[!is.na(exp)],
                          method = method, ...)
        applied[!is.na(exp_cor)] <- doQmap(exp_cor[!is.na(exp_cor)],
                                           adjust, ...)
      },
      error = function(error_message) {
        return(applied)
      })
    } else {
      adjust <- fitQmap(as.vector(obs), as.vector(exp), method = method, ...)
      applied <- doQmap(as.vector(exp_cor), adjust, ...)
    }
    if (length(applied) == 1) {
      applied <- matrix(applied, nrow = nrow(exp_cor), ncol = ncol(exp_cor))
    } else {
      applied <- array(applied, dim = dim(exp_cor))
    }
    dim(applied) <- dim(exp_cor)
  }
  return(applied)
}

fitQmap <- function (obs, mod, method = c("PTF", "DIST", "RQUANT", "QUANT", 
                                          "SSPLIN"), ...) 
{
  # ffun <- match.fun(paste("fitQmap", method, sep = ""))
  ffun <- match.fun(qmap::fitQmapQUANT)

  ffun(obs, mod, ...)
  # qmap::fitQmapQUANT(obs,mod,...)
}

doQmap <- function (x, fobj, ...) 
  {
    cc <- class(fobj)
    ffun <- substring(cc, 4, nchar(cc))
    # ffun <- paste("do", ffun, sep = "")
    # test <- sapply(paste("qmap::", ffun, sep = ""), exists, mode = "function")
    test <- exists("doQmapQUANT", envir = asNamespace("qmap"), mode = "function")
    ffun_name <- paste("do", ffun, sep = "")
    ffun <- get(ffun_name, envir = asNamespace("qmap"))
    
    if (all(test)) {
      # ffun <- match.fun(paste("qmap::", ffun, sep = ""))
      ffun <- match.fun(ffun)
    }
    else {
      stop("doQmap not defined for class(fobj) ==", class(fobj))
    }
    ffun(x, fobj, ...)
}

