
library(multiApply)
library(s2dv)
library(ClimProjDiags)
library(modifiedmk)

Trend_edited <- function(data, time_dim = 'year', member_dim = NULL, alpha = 0.05, na.rm = FALSE, f_NAs = 1, ncores = 1){
  
  
  ## Ensemble mean
  if (!is.null(member_dim)){
    data <- multiApply::Apply(data = data, target_dims = member_dim, fun = mean, na.rm = na.rm, ncores = ncores)$output1
  }
  
  ## Trend and significance
  output <- multiApply::Apply(data = data, target_dims = time_dim, fun = .Trend_edited, 
                              f_NAs = f_NAs, alpha = alpha, ncores = ncores)
  return(output)
}

.Trend_edited <- function(data, f_NAs, alpha){

  good_values <- !is.na(data) ## Criterion for NAs (is less than f_NAs is missing)
  if (f_NAs <= sum(good_values)/length(data)){
    
    data <- data[good_values]
    
    ## Trend
    trend <- s2dv:::.Trend(x = data, interval = 1, polydeg = 1, conf = FALSE, alpha = alpha, pval = FALSE, sign = FALSE)
    trend <- trend$trend[2] ## keeping the trend (removing the intercept)
    
    ## Significance test taking into account the autocorrelation (Hamed and Rao, 1998)
    # Hamed, K. H. and Rao, A. R. (1998). A modified Mann-Kendall trend test for autocorrelated data.
    # Journal of Hydrology, 204(1–4): 182–196. <doi:10.1016/S0022-1694(97)00125-X>
    sign <- .mmkh_fixed(x = as.vector(data), ci = 1-alpha)
    sign <- sign[['new P-value']] <= alpha/2 ## two-sided test
    
    return(list(trend = trend, sign = sign))
    
  } else {
    
    return(list(trend = NA, sign = NA))
    
  }
  
}

## modifiedmk::mmkh fails when evaluating "if(ro[i] > sig || ro[i] < -sig)" and ro[i] is NaN
## I have added is.finite(ro[i]) to the condition
.mmkh_fixed <-function(x, ci=0.95) {
  # Initialize the test parameters
  options(scipen = 999)
  # Time series vector
  x = x
  # Modified Z statistic after variance correction by Hamed and Rao (1998) method
  z = NULL
  # Original Z statistic for Mann-Kendall test before variance correction
  z0 = NULL
  # Modified Z statistic after variance correction by Hamed and Rao (1998) method
  pval = NULL
  # Original p-value for Mann-Kendall test before variance correction
  pval0 = NULL
  # Initialize Mann-Kendall S statistic
  S = 0
  # Initialize Mann-Kendall Tau
  Tau = NULL
  # Correction factor n/n* value
  essf = NULL
  # Confidance interval
  ci = ci
  
  # To test whether the data is in vector format
  
  if (is.vector(x) == FALSE) {
    stop("Input data must be a vector")
  }
  
  # To test whether the data values are finite numbers and attempting to eliminate non-finite numbers
  
  if (any(is.finite(x) == FALSE)) {
    x[-c(which(is.finite(x) == FALSE))] -> x
    warning("The input vector contains non-finite numbers. An attempt was made to remove them")
  }
  
  n <- length(x)
  
  #Specify minimum input vector length
  if (n < 3) {
    stop("Input vector must contain at least three values")
  }
  
  # Calculating Sen's slope
  rep(NA, n * (n - 1)/2) -> V
  k = 0
  for (i in 1:(n-1)) {
    for (j in (i+1):n) {
      k = k+1
      V[k] = (x[j]-x[i])/(j-i)
    }
  }
  median(V,na.rm=TRUE)->slp
  
  # Calculating trend-free Series
  
  t=1:length(x)
  xn<-(x[1:n])-((slp)*(t))
  
  
  # Calculating Mann-Kendall S statistic
  
  for (i in 1:(n-1)) {
    for (j in (i+1):n) {
      S = S + sign(x[j]-x[i])
    }
  }
  
  # Calculating autocorrelation function of the ranks of observations (ro)
  
  acf(rank(xn), lag.max=(n-1), plot=FALSE)$acf[-1] -> ro
  
  # Calculating significant autocorrelation at given confidance interval (rof)
  
  qnorm((1+ci)/2)/sqrt(n) -> sig
  rep(NA,length(ro)) -> rof
  for (i in 1:(length(ro))) {
    if (is.finite(ro[i]) && (ro[i] > sig || ro[i] < -sig)) {
      rof[i] <- ro[i]
    } else {
      rof[i] = 0
    }
  }
  
  # Calculating 2/(n*(n-1)*(n-2))
  
  2 / (n*(n-1)*(n-2)) -> cte
  
  # Calculating sum(((n-i)*(n-i-1)*(n-i-2)*rof[i]
  
  ess=0
  for (i in 1:(n-1)) {
    ess = ess + (n-i)*(n-i-1)*(n-i-2)*rof[i]
  }
  
  # Calculating variance correction factor (n/n*) as per Hamed and Rao (1998)
  
  essf = 1 + ess*cte
  
  # Calculating Mann-Kendall Variance before correction (Var(s))
  
  var.S = n*(n-1)*(2*n+5)*(1/18)
  if(length(unique(x)) < n) {
    unique(x) -> aux
    for (i in 1:length(aux)) {
      length(which(x == aux[i])) -> tie
      if (tie > 1) {
        var.S = var.S - tie*(tie-1)*(2*tie+5)*(1/18)
      }
    }
  }
  
  # Calculating new variance Var(s)*=(Var(s))*(n/n*) as per Hamed and Rao (1998)
  
  VS = var.S * essf
  
  # Calculating Z statistic values before and after variance correction
  
  if (S == 0) {
    z = 0
    z0 = 0
  }else
    if (S > 0) {
      z = (S-1)/sqrt(VS)
      z0 = (S-1)/sqrt(var.S)
    } else {
      z = (S+1)/sqrt(VS)
      z0 = (S+1)/sqrt(var.S)
    }
  
  # Calculating p-value before and after variance correction
  
  pval = 2*pnorm(-abs(z))
  pval0 = 2*pnorm(-abs(z0))
  
  # Calculating Kendall's Tau
  
  Tau = S/(.5*n*(n-1))
  
  
  return(c("Corrected Zc" = z,
           "new P-value" = pval,
           "N/N*" = essf,
           "Original Z" = z0,
           "old P.value" = pval0,
           "Tau" = Tau,
           "Sen's slope" = slp,
           "old.variance"=var.S,
           "new.variance"= VS))
}
