
library(shiny)
library(tibble)
library(dplyr)
library(shinyjs)
library(shinyauthr)
# library(shinythemes)

user_base <- tibble::tibble(
  user = "smoreno",
  password = "climateservices"
  #  permissions = c("admin", "standard"),
  #  name = c("User One", "User Two")
)

ui <- fluidPage(
  title = 'Sara Moreno - Shiny App',
  div(class = "pull-right", shinyauthr::logoutUI(id = "logout")),
  shinyauthr::loginUI(id = "login", 
                      title = 'Please log in',
                      user_title = "User Name",
                      pass_title = "Password",
                      login_title = "Log in",
                      error_message = "Invalid username or password!",
                      additional_ui = HTML('<center><H4> Please send an email to <a href="mailto:sara.moreno@bsc.es">sara.moreno@bsc.es</a> asking for the credentials 
                                           or check the <a href="https://earth.bsc.es/wiki/doku.php?id=working_groups:earth_system_services:esscredentials">wiki</a> </H4></center>'),
                      cookie_expiry = 7),
  uiOutput("sidebarpanel"),
  plotOutput("figure")
)

server <- function(input, output) {
  
  ## Login
  credentials <- shinyauthr::loginServer(
    id = "login",
    data = user_base,
    user_col = user,
    pwd_col = password,
    sodium_hashed = FALSE,
    log_out = reactive(logout_init())
  )
  
  ## Logout to hide
  logout_init <- shinyauthr::logoutServer(
    id = "logout",
    active = reactive(credentials()$user_auth)
  )
  
  ## User interface
  output$sidebarpanel <- renderUI({
    
    req(credentials()$user_auth)
    
    tabsetPanel(
      tabPanel('Solar capacity factors', fluid = TRUE,
               sidebarLayout(
                 sidebarPanel(
                   width = 2,
                   checkboxGroupInput(inputId = 'decadal2b_model', label = 'Model', 
                                      choices = c('Obs'='obs',
                                                  'Multi-model-mean',
                                                  'BCC-CSM2-MR'='dcpp_BCC-CSM2-MR',
                                                  'CanESM5'='dcpp_CanESM5',
                                                  'CMCC-CM2-SR5'='dcpp_CMCC-CM2-SR5',
                                                  'EC-Earth3-i1'='dcpp_EC-Earth3-i1',
                                                  'EC-Earth3-i2'='dcpp_EC-Earth3-i2',
                                                  'EC-Earth3-i4'='dcpp_EC-Earth3-i4',
                                                  'HadGEM3-GC3.1-MM'='dcpp_HadGEM3-GC3.1-MM',
                                                  'IPSL-CM6A-LR'='dcpp_IPSL-CMG1-LR',
                                                  'MIROC6'='dcpp_MIROC6',
                                                  'MPI-ESM1.2-HR'='dcpp_MPI-ESM1.2-HR',
                                                  'MRI-ESM2-0'='dcpp_MRI-ESM2-0',
                                                  'NorCPM1-i1'='dcpp_NorCPM1-i1',
                                                  'NorCPM1-i2'='dcpp_NorCPM1-i2'),
                                      inline = FALSE, width = '100%'),
                   checkboxGroupInput(inputId = 'decadal2b_result', label = 'Results', 
                                      choices = list('capacity_factors' = 'capacity',
                                                     'days_under_10' = 'days_under_10',
                                                     'days_over_25' = 'days_over_25',
                                                     'ACC' = 'acc',
                                                     'CRPSS' = 'crpss'),
                                      inline = FALSE, width = '100%'),
                   sliderInput(inputId = 'decadal2b_size', label = 'Plot size (%)',
                               min = 0, max = 200, step = 1, value = 100,
                               round = TRUE, animate = FALSE, ticks = TRUE)
                 ),
                 mainPanel(
                   uiOutput(outputId = 'decadal2b_plots')
                 )
               )
      ),
      
      tabPanel('Info', fluid = TRUE,
               mainPanel(HTML('
                            <H4> Forecast systems </H4>
                            <ul>
                            <li>BCC-CSM2-MR: 8 members</li>
                            <li>CanESM5: 20 members</li>
                            <li>CMCC-CM2-SR5: 10 members</li>
                            <li>EC-Earth3-i1: 10 members</li>
                            <li>EC-Earth3-i2: 5 members</li>
                            <li>EC-Earth3-i4: 10 members</li>
                            <li>HadGEM3-GC3.1-MM: 10 members</li>
                            <li>IPSL-CM6A-LR: 10 members</li>
                            <li>MIROC6: 10 members</li>
                            <li>MPI-ESM1.2-HR: 10 members</li>
                            <li>MRI-ESM2-0: 10 members</li>
                            <li>NorCPM1-i1: 10 members</li>
                            <li>NorCPM1-i2: 10 members</li>
                            </ul>
                            <H4> Variables</H4>
                            <ul>
                            <li>solar: computed with rsds and tas</li>
                            <li>sfcWind: surface wind velocity</li>
                            </ul>
                            <H4> Reference datasets </H4>
                            <ul>
                            <li>ERA5land</li>
                            </ul>
                            '
               ))
      )
    )
    
  })
  
  #######################################################################################
  ################################## DECADAL2b TAB ######################################
  #######################################################################################
  
  decadal2b_n_plots <- reactive({
    length(input$decadal2b_model) * length(input$decadal2b_result)
  })
  decadal2b_n_models <- reactive({
    length(input$decadal2b_model)
  })
  decadal2b_n_results <- reactive({
    length(input$decadal2b_result)
  })
  
  decadal2b_plots <- reactive({
    x = NULL
    for (a in 1:decadal2b_n_models()){
      for (b in 1:decadal2b_n_results()){
        x[[length(x)+1]] <- paste0('plot_decadal2b_',a,'-',b)
      }
    }
    return(x)
  })
  
  output$decadal2b_plots <- renderUI({
    plot_output_list <- lapply(1:decadal2b_n_plots(), function(i) {
      plotname <- decadal2b_plots()[i]
      imageOutput(outputId = plotname, 
                  width = paste0(input$decadal2b_size,'%'), height = paste0(input$decadal2b_size,'%'))
    })
    do.call(tagList, plot_output_list)
  })
  
  observeEvent(decadal2b_plots(), {
    for (a in 1:decadal2b_n_models()) { 
      my_a <- a
      for (b in 1:decadal2b_n_results()) { 
        my_b <- b
        local({
          plotname <- paste0('plot_decadal2b_', my_a, '-', my_b)
          filename <- paste0('/esarchive/scratch/smoreno/capacity_outputs/figures_capacity/seasons_capacity_factors_', input$decadal2b_result[my_b],
                             '_', input$decadal2b_model[my_a], '.png')
          print(filename)
          output[[plotname]] <- renderImage({
            return(list(
              src = filename,
              contentType = 'image/png',
              width = paste0(input$decadal2b_size, '%'),
              height = paste0(input$decadal2b_size, '%'),
              alt = 'Please select one or more options for each category on the left panel'
            ))
          }, deleteFile = FALSE)
        })
      }
    }
  })
}

shinyApp(ui = ui, server = server)
